import numpy

from paysage import backends as be
from paysage import batch
from paysage import layers
from paysage.models import BoltzmannMachine
from paysage import fit
from paysage import samplers
from paysage import optimizers
from paysage import schedules

import pytest

def test_gaussian_1D_1mode_train():
    # create some example data
    num = 10000
    mu = 3
    sigma = 1
    samples = be.randn((num, 1)) * sigma + mu

    # set up the reader to get minibatches
    batch_size = 100
    samples_train, samples_validate = batch.split_tensor(samples, 0.9)
    data = batch.Batch({'train': batch.InMemoryTable(samples_train, batch_size),
                        'validate': batch.InMemoryTable(samples_validate, batch_size)})

    # parameters
    learning_rate = schedules.PowerLawDecay(initial=0.1, coefficient=0.1)
    mc_steps = 1
    num_epochs = 10
    num_sample_steps = 100

    # set up the model and initialize the parameters
    vis_layer = layers.GaussianLayer(1)
    hid_layer = layers.OneHotLayer(1)

    rbm = BoltzmannMachine([vis_layer, hid_layer])
    rbm.initialize(data, method='hinton')

    # modify the parameters to shift the initialized model from the data
    # this forces it to train
    rbm.layers[0].params = layers.ParamsGaussian(rbm.layers[0].params.loc - 3,
                                                 rbm.layers[0].params.log_var - 1)

    # set up the optimizer and the fit method
    opt = optimizers.ADAM(stepsize=learning_rate)
    cd = fit.SGD(rbm, data)

    # fit the model
    print('training with persistent contrastive divergence')
    cd.train(opt, num_epochs, method=fit.pcd, mcsteps=mc_steps)

    # sample data from the trained model
    model_state = \
        samplers.SequentialMC.generate_fantasy_state(rbm, num, num_sample_steps)
    pts_trained = model_state[0]

    percent_error = 10
    mu_trained = be.mean(pts_trained)
    assert numpy.abs(mu_trained / mu - 1) < (percent_error/100)

    sigma_trained = numpy.sqrt(be.var(pts_trained))
    assert numpy.abs(sigma_trained / sigma - 1) < (percent_error/100)

def test_gaussian_1D_1mode_stationary():
    # create some example data
    num = 10000
    mu = 3
    sigma = 1
    samples = be.randn((num, 1)) * sigma + mu

    # set up the reader to get minibatches
    batch_size = 100
    samples_train, samples_validate = batch.split_tensor(samples, 0.9)
    data = batch.Batch({'train': batch.InMemoryTable(samples_train, batch_size),
                        'validate': batch.InMemoryTable(samples_validate, batch_size)})

    # parameters
    learning_rate = schedules.PowerLawDecay(initial=0.1, coefficient=0.1)
    mc_steps = 1
    num_epochs = 10
    num_sample_steps = 100

    # set up the model and initialize the parameters
    vis_layer = layers.GaussianLayer(1)
    hid_layer = layers.OneHotLayer(1)

    rbm = BoltzmannMachine([vis_layer, hid_layer])
    # keep the parameters as initialized.
    # the model should not change much in training.
    rbm.initialize(data, method='hinton')

    # set up the optimizer and the fit method
    opt = optimizers.ADAM(stepsize=learning_rate)
    cd = fit.SGD(rbm, data)

    # fit the model
    print('training with persistent contrastive divergence')
    cd.train(opt, num_epochs, method=fit.pcd, mcsteps=mc_steps)

    # sample data from the trained model
    model_state = samplers.SequentialMC.generate_fantasy_state(rbm, num,
                                                               num_sample_steps)
    pts_trained = model_state[0]

    percent_error = 10
    mu_trained = be.mean(pts_trained)
    assert numpy.abs(mu_trained / mu - 1) < (percent_error/100)

    sigma_trained = numpy.sqrt(be.var(pts_trained))
    assert numpy.abs(sigma_trained / sigma - 1) < (percent_error/100)

if __name__ == "__main__":
    pytest.main([__file__])



import sys

from . import backends as be


class Penalty(object):
    """
    Base penalty class.
    Derived classes should define `value` and `grad` functions.

    """
    def __init__(self, penalty, slice_tuple = (slice(None,None,None),)):
        """
        Create a base Penalty object.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            Penalty

        """
        self.penalty = penalty
        self.slice_tuple = slice_tuple

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        raise NotImplementedError

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        raise NotImplementedError

    def get_config(self):
        """
        Returns a config for the penalty.

        Args:
            None

        Returns:
            dict

        """
        return {"name": self.__class__.__name__,
                "penalty": self.penalty,
                "slice": self.slice_tuple
               }


class trivial_penalty(Penalty):
    """A penalty that does nothing."""
    def __init__(self, penalty_unused=0, slice_tuple_unused=(slice(None,None,None),)):
        """
        Create a base trivial penalty.

        Args:
            penalty_unused (float): strength of the penalty
            slice_tuple_unused (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            trivial_penalty

        """
        super().__init__(penalty_unused)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return 0.0

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        return be.zeros_like(tensor)


class l2_penalty(Penalty):
    """
    An L2 penalty encourages small values of the parameters.
    Also known as a "ridge" penalty.

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create an l2 penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            l2_penalty

        """
        super().__init__(penalty, slice_tuple)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return 0.5 * self.penalty * be.tsum(tensor[self.slice_tuple]**2)

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = self.penalty * tensor[self.slice_tuple]
        return retval


class l1_penalty(Penalty):
    """
    An l1 penalty encourages small values of the parameters,
    and tends to produce solutions that are more sparse than an l2 penalty.
    Also known as "lasso".

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create an l1 penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            l1_penalty

        """
        super().__init__(penalty, slice_tuple)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return self.penalty * be.tsum(be.tabs(tensor[self.slice_tuple]))

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = self.penalty * be.sign(tensor[self.slice_tuple])
        return retval


class exp_l2_penalty(Penalty):
    """
    Puts an l2 penalty on the exponentiated parameters.
    Useful when the parameters are represented in a logarithmic space.
    For example, encouraging the variances of a GaussianLayer to be small.

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create an exp l2 penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            exp_l2_penalty

        """
        super().__init__(penalty, slice_tuple)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return 0.5 * self.penalty * be.tsum(be.exp(2.0*tensor[self.slice_tuple]))

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = self.penalty * be.exp(2.0*tensor[self.slice_tuple])
        return retval


class l1_adaptive_decay_penalty_2(Penalty):
    """
    Modified form of the l1 penalty which regularizes more for weight rows
    with larger coupling to target layer by way of a quadratic power

    Tubiana, J., Monasson, R. (2017)
    Emergence of Compositional Representations in Restricted Boltzmann Machines,
    PRL 118, 138301 (2017), Supplemental Material I.D

    Note: expects to operate on a tensor with two degrees of freedom (eg. a weight matrix)

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create an adaptive l1 penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            l1_adaptive_decay_penalty_2

        """
        super().__init__(penalty, slice_tuple)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return 0.5 * self.penalty * \
               be.tsum(be.square(be.tsum(be.tabs(tensor[self.slice_tuple]),axis=0)))

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = self.penalty * \
               be.multiply(be.tsum(be.tabs(tensor[self.slice_tuple]),axis=0),
                           be.sign(tensor[self.slice_tuple]))
        return retval


class log_penalty(Penalty):
    """A logarithmic penalty forces the parameters to be positive."""
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create a log penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            log_penalty

        """
        super().__init__(penalty, slice_tuple)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return -self.penalty * be.tsum(be.log(tensor[self.slice_tuple]))

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = -self.penalty / tensor[self.slice_tuple]
        return retval


class logdet_penalty(Penalty):
    """
    Penalty acts on the logarithm of the determinant of a matrix.
    Discourages the matrix from becoming singular.

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create a logdet_penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            logdet_penalty

        """
        super().__init__(penalty, slice_tuple)
        assert len(slice_tuple) < 2, "logdet doesn't work with multiindex slices"

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        J = be.dot(be.transpose(tensor[self.slice_tuple]), tensor[self.slice_tuple])
        return -self.penalty * be.logdet(J)

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        retval = be.ones_like(tensor)
        retval[self.slice_tuple] = -2 * self.penalty * be.transpose(be.pinv(tensor[self.slice_tuple]))
        return retval


class log_norm(Penalty):
    """
    An RBM with n visible units and m hidden units has an (n, m) weight matrix.
    The log norm penalty discourages any of the m columns of the weight matrix
    from having zero norm.

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),)):
        """
        Create an log_norm.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied

        Returns:
            log_norm

        """
        super().__init__(penalty, slice_tuple)

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        return -self.penalty * 0.5*be.tsum(be.log(be.square(be.norm(tensor[self.slice_tuple], axis=0))))

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        tmp = be.square(be.norm(tensor[self.slice_tuple], axis=0))
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = -self.penalty * be.divide(tmp, tensor[self.slice_tuple])
        return retval


class l2_norm(Penalty):
    """
    An RBM with n visible units and m hidden units has an (n, m) weight matrix.
    The l2 norm penalty encourages the columns of the weight matrix to have
    norms that are close to the target value.

    "On Training Deep Boltzmann Machines"
    by Guillaume Desjardins, Aaron Courville, Yoshua Bengio
    http://arxiv.org/pdf/1203.4416.pdf

    """
    def __init__(self, penalty, slice_tuple=(slice(None,None,None),), target=1):
        """
        Create an l2_norm penalty.

        Args:
            penalty (float): strength of the penalty
            slice_tuple (slice): list of slices that define the parts of the
                tensor to which the penalty will be applied
            target (optional; float): the shrinkage target

        Returns:
            l2_norm

        """
        super().__init__(penalty, slice_tuple)
        self.target = target

    def value(self, tensor):
        """
        The value of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            float

        """
        norms = be.norm(tensor[self.slice_tuple], axis=0)
        return 0.5 * self.penalty * be.tsum(be.square(norms - self.target))

    def grad(self, tensor):
        """
        The value of the gradient of the penalty function on a tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        norms = be.EPSILON + be.norm(tensor[self.slice_tuple], axis=0)
        retval = be.zeros_like(tensor)
        retval[self.slice_tuple] = \
            self.penalty * (self.target- 1 / norms) * tensor[self.slice_tuple]
        return retval


# ----- FUNCTIONS ----- #

def from_config(config):
    """
    Builds an instance from a config.

    Args:
        config (List or dict)

    Returns:
        Penalty

    """
    empty_slice = (slice(None,None,None),)
    configs = be.force_list(config)
    pens = [getattr(sys.modules[__name__], c["name"]) for c in configs]
    penalties = \
        [pens[i](configs[i]["penalty"], be.maybe_key(configs[i],"slice", empty_slice))
         for i in range(len(pens))]
    if len(penalties) == 1:
        return penalties[0]
    return PenaltySum(penalties)


# ----- ALIASES ----- #

ridge = l2_penalty
lasso = l1_penalty



from paysage import backends as be
from paysage import layers
from paysage.models import BoltzmannMachine
from paysage.models import gradient_util as gu
from paysage.models.state import StateTAP
import pytest
from copy import deepcopy
from cytoolz import partial
import math

# ----- Functional Programs with Gradients ----- #

def test_zero_grad():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with zeros
    gu.zero_grad(rbm)

def test_random_grad():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    gu.random_grad(rbm)

def test_grad_accumulate():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad = gu.random_grad(rbm)
    gu.grad_accumulate(be.norm, grad)

def test_grad_apply():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad = gu.random_grad(rbm)
    gu.grad_apply(be.square, grad)

def test_grad_apply_():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad = gu.random_grad(rbm)
    gu.grad_apply_(be.square, grad)

def test_grad_mapzip():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad_1 = gu.random_grad(rbm)
    grad_2 = gu.random_grad(rbm)
    gu.grad_mapzip(be.add, grad_1, grad_2)

def test_grad_mapzip_():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad_1 = gu.random_grad(rbm)
    grad_2 = gu.random_grad(rbm)
    gu.grad_mapzip_(be.add_, grad_1, grad_2)

def test_grad_rms():
    num_visible_units = 100
    num_hidden_units = 50

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad = gu.zero_grad(rbm)
    mag = gu.grad_rms(grad)
    assert mag == 0

def test_grad_norm():
    num_visible_units = 1000
    num_hidden_units = 1000

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad = gu.random_grad(rbm)
    nrm = gu.grad_norm(grad)
    assert nrm > math.sqrt(be.float_scalar(num_hidden_units + \
        num_visible_units + num_visible_units*num_hidden_units)/3) - 1
    assert nrm < math.sqrt(be.float_scalar(num_hidden_units + \
        num_visible_units + num_visible_units*num_hidden_units)/3) + 1

def test_grad_normalize_():
    num_visible_units = 10
    num_hidden_units = 10

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # create a gradient object filled with random numbers
    grad = gu.random_grad(rbm)
    gu.grad_normalize_(grad)
    nrm = gu.grad_norm(grad)
    assert nrm > 1-1e-6
    assert nrm < 1+1e-6


# ----- Layer Methods ----- #

def test_bernoulli_conditional_params():
    num_visible_units = 100
    num_hidden_units = 50
    batch_size = 25

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # randomly set the intrinsic model parameters
    a = be.randn((num_visible_units,))
    b = be.randn((num_hidden_units,))
    W = be.randn((num_visible_units, num_hidden_units))

    rbm.layers[0].params.loc[:] = a
    rbm.layers[1].params.loc[:] = b
    rbm.connections[0].weights.params.matrix[:] = W

    # generate a random batch of data
    vdata = rbm.layers[0].random((batch_size, num_visible_units))
    hdata = rbm.layers[1].random((batch_size, num_hidden_units))

    # compute conditional parameters
    hidden_field = be.dot(vdata, W) # (batch_size, num_hidden_units)
    hidden_field += b

    visible_field = be.dot(hdata, be.transpose(W)) # (batch_size, num_visible_units)
    visible_field += a

    # compute conditional parameters with layer funcitons
    hidden_field_layer = rbm.layers[1].conditional_params(
        [vdata], [rbm.connections[0].W()])
    visible_field_layer = rbm.layers[0].conditional_params(
        [hdata], [rbm.connections[0].W(trans=True)])

    assert be.allclose(hidden_field, hidden_field_layer), \
    "hidden field wrong in bernoulli-bernoulli rbm"

    assert be.allclose(visible_field, visible_field_layer), \
    "visible field wrong in bernoulli-bernoulli rbm"


def test_bernoulli_derivatives():
    num_visible_units = 100
    num_hidden_units = 50
    batch_size = 25

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.BernoulliLayer(num_visible_units)
    hid_layer = layers.BernoulliLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # randomly set the intrinsic model parameters
    a = be.randn((num_visible_units,))
    b = be.randn((num_hidden_units,))
    W = be.randn((num_visible_units, num_hidden_units))

    rbm.layers[0].params.loc[:] = a
    rbm.layers[1].params.loc[:] = b
    rbm.connections[0].weights.params.matrix[:] = W

    # generate a random batch of data
    vdata = rbm.layers[0].random((batch_size, num_visible_units))
    vdata_scaled = rbm.layers[0].rescale(vdata)

    # compute the conditional mean of the hidden layer
    hid_mean = rbm.layers[1].conditional_mean([vdata], [rbm.connections[0].W()])
    hid_mean_scaled = rbm.layers[1].rescale(hid_mean)

    # compute the derivatives
    d_visible_loc = -be.mean(vdata, axis=0)
    d_hidden_loc = -be.mean(hid_mean_scaled, axis=0)
    d_W = -be.batch_outer(vdata, hid_mean_scaled) / len(vdata)

    # compute the derivatives using the layer functions
    vis_derivs = rbm.layers[0].derivatives(vdata, [hid_mean_scaled],
                                            [rbm.connections[0].W(trans=True)])

    hid_derivs = rbm.layers[1].derivatives(hid_mean, [vdata_scaled],
                                          [rbm.connections[0].W()])

    weight_derivs = rbm.connections[0].weights.derivatives(vdata, hid_mean_scaled)

    # compute simple weighted derivatives using the layer functions
    scale = 2
    scale_func = partial(be.multiply, be.float_scalar(scale))
    vis_derivs_scaled = rbm.layers[0].derivatives(vdata, [hid_mean_scaled],
                        [rbm.connections[0].W(trans=True)], weighting_function=scale_func)

    hid_derivs_scaled = rbm.layers[1].derivatives(hid_mean, [vdata_scaled],
                        [rbm.connections[0].W()], weighting_function=scale_func)

    weight_derivs_scaled = rbm.connections[0].weights.derivatives(vdata, hid_mean_scaled,
                                                weighting_function=scale_func)

    assert be.allclose(d_visible_loc, vis_derivs[0].loc), \
    "derivative of visible loc wrong in bernoulli-bernoulli rbm"

    assert be.allclose(d_hidden_loc, hid_derivs[0].loc), \
    "derivative of hidden loc wrong in bernoulli-bernoulli rbm"

    assert be.allclose(d_W, weight_derivs[0].matrix), \
    "derivative of weights wrong in bernoulli-bernoulli rbm"

    assert be.allclose(scale * d_visible_loc, vis_derivs_scaled[0].loc), \
    "weighted derivative of visible loc wrong in bernoulli-bernoulli rbm"

    assert be.allclose(scale * d_hidden_loc, hid_derivs_scaled[0].loc), \
    "weighted derivative of hidden loc wrong in bernoulli-bernoulli rbm"

    assert be.allclose(scale * d_W, weight_derivs_scaled[0].matrix), \
    "weighted derivative of weights wrong in bernoulli-bernoulli rbm"


def test_bernoulli_GFE_entropy_gradient():
    num_units = 5

    lay = layers.BernoulliLayer(num_units)
    lay.params.loc[:] = be.rand_like(lay.params.loc)

    from cytoolz import compose
    sum_square = compose(be.tsum, be.square)

    for itr in range(10):
        lr = 0.1
        gogogo = True
        mag = lay.get_random_magnetization()
        lms = lay.lagrange_multipliers_analytic(mag)
        entropy = lay.TAP_entropy(mag)

        grad = lay.TAP_magnetization_grad(mag, [], [], [])
        grad_mag = math.sqrt(be.float_scalar(be.accumulate(sum_square, grad)))
        normit = partial(be.tmul_, be.float_scalar(1.0/grad_mag))
        be.apply_(normit, grad)
        rand_grad = lay.get_random_magnetization()
        grad_mag = math.sqrt(be.float_scalar(be.accumulate(sum_square, rand_grad)))
        normit = partial(be.tmul_, be.float_scalar(1.0/grad_mag))
        be.apply_(normit, rand_grad)
        while gogogo:
            cop1_mag = deepcopy(mag)
            cop1_lms = deepcopy(lms)
            cop2_mag = deepcopy(mag)
            cop2_lms = deepcopy(lms)

            cop1_mag.mean[:] = mag.mean + lr * grad.mean
            cop2_mag.mean[:] = mag.mean + lr * rand_grad.mean
            lay.clip_magnetization_(cop1_mag)
            lay.clip_magnetization_(cop2_mag)
            cop1_lms = lay.lagrange_multipliers_analytic(cop1_mag)
            cop2_lms = lay.lagrange_multipliers_analytic(cop2_mag)

            entropy_1 = lay.TAP_entropy(cop1_mag)
            entropy_2 = lay.TAP_entropy(cop2_mag)

            regress = entropy_1 - entropy_2 < 0.0
            #print(itr, "[",lr, "] ", entropy, entropy_1, entropy_2, regress)
            if regress:
                #print(grad, rand_grad)
                if lr < 1e-6:
                    assert False,\
                    "Bernoulli GFE magnetization gradient is wrong"
                    break
                else:
                    lr *= 0.5
            else:
                break

def test_gaussian_GFE_entropy_gradient():
    num_units = 5
    lay = layers.GaussianLayer(num_units)

    lay.params.loc[:] = be.rand_like(lay.params.loc)
    lay.params.log_var[:] = be.randn(be.shape(lay.params.loc))

    from cytoolz import compose
    sum_square = compose(be.tsum, be.square)

    for itr in range(10):
        mag = lay.get_random_magnetization()
        lms = lay.lagrange_multipliers_analytic(mag)
        entropy = lay.TAP_entropy(mag)
        lr = 0.001
        gogogo = True
        grad = lay.TAP_magnetization_grad(mag, [], [], [])
        grad_mag = math.sqrt(be.float_scalar(be.accumulate(sum_square, grad)))
        normit = partial(be.tmul_, be.float_scalar(1.0/grad_mag))
        be.apply_(normit, grad)
        rand_grad = lay.get_random_magnetization()
        grad_mag = math.sqrt(be.float_scalar(be.accumulate(sum_square, rand_grad)))
        normit = partial(be.tmul_, be.float_scalar(1.0/grad_mag))
        be.apply_(normit, rand_grad)
        while gogogo:
            cop1_mag = deepcopy(mag)
            cop1_lms = deepcopy(lms)
            cop2_mag = deepcopy(mag)
            cop2_lms = deepcopy(lms)

            cop1_mag.mean[:] = mag.mean + lr * grad.mean
            cop2_mag.mean[:] = mag.mean + lr * rand_grad.mean
            cop1_mag.variance[:] = mag.variance + lr * grad.variance
            cop2_mag.variance[:] = mag.variance + lr * rand_grad.variance
            lay.clip_magnetization_(cop1_mag)
            lay.clip_magnetization_(cop2_mag)
            cop1_lms = lay.lagrange_multipliers_analytic(cop1_mag)
            cop2_lms = lay.lagrange_multipliers_analytic(cop2_mag)

            entropy_1 = lay.TAP_entropy(cop1_mag)
            entropy_2 = lay.TAP_entropy(cop2_mag)

            regress = entropy_1 - entropy_2 < 0.0
            #print(itr, "[",lr, "] ", entropy, entropy_1, entropy_2, regress)
            if regress:
                #print(grad, rand_grad)
                if lr < 1e-6:
                    assert False,\
                    "Gaussian GFE magnetization gradient is wrong"
                    break
                else:
                    lr *= 0.5
            else:
                break

def test_bernoulli_GFE_derivatives():
    # Tests that the GFE derivative update increases GFE versus 100
    # random update vectors
    num_units = 5

    layer_1 = layers.BernoulliLayer(num_units)
    layer_2 = layers.BernoulliLayer(num_units)
    layer_3 = layers.BernoulliLayer(num_units)
    rbm = BoltzmannMachine([layer_1, layer_2, layer_3])

    for i in range(len(rbm.connections)):
        rbm.connections[i].weights.params.matrix[:] = \
        0.01 * be.randn(rbm.connections[i].shape)

    for lay in rbm.layers:
        lay.params.loc[:] = be.rand_like(lay.params.loc)

    state, cop1_GFE = rbm.compute_StateTAP(init_lr=0.1, tol=1e-7, max_iters=50)
    grad = rbm._grad_gibbs_free_energy(state)
    gu.grad_normalize_(grad)

    for i in range(100):
        lr = 1.0
        gogogo = True
        random_grad = gu.random_grad(rbm)
        gu.grad_normalize_(random_grad)
        while gogogo:
            cop1 = deepcopy(rbm)
            lr_mul = partial(be.tmul, lr)

            cop1.parameter_update(gu.grad_apply(lr_mul, grad))
            cop1_state, cop1_GFE = cop1.compute_StateTAP(init_lr=0.1, tol=1e-7, max_iters=50)

            cop2 = deepcopy(rbm)
            cop2.parameter_update(gu.grad_apply(lr_mul, random_grad))
            cop2_state, cop2_GFE = cop2.compute_StateTAP(init_lr=0.1, tol=1e-7, max_iters=50)

            regress = cop2_GFE - cop1_GFE < 0.0

            if regress:
                if lr < 1e-6:
                    assert False, \
                    "TAP FE gradient is not working properly for Bernoulli models"
                    break
                else:
                    lr *= 0.5
            else:
                break

def test_gaussian_GFE_derivatives_gradient_descent():
    num_units = 5

    layer_1 = layers.GaussianLayer(num_units)
    layer_2 = layers.BernoulliLayer(num_units)

    rbm = BoltzmannMachine([layer_1, layer_2])

    for i in range(len(rbm.connections)):
        rbm.connections[i].weights.params.matrix[:] = \
        0.01 * be.randn(rbm.connections[i].shape)

    for lay in rbm.layers:
        lay.params.loc[:] = be.rand_like(lay.params.loc)

    state, GFE = rbm.compute_StateTAP(use_GD=False, tol=1e-7, max_iters=50)
    grad = rbm._grad_gibbs_free_energy(state)
    gu.grad_normalize_(grad)

    for i in range(100):
        lr = 0.001
        gogogo = True
        random_grad = gu.random_grad(rbm)
        gu.grad_normalize_(random_grad)
        while gogogo:
            cop1 = deepcopy(rbm)
            lr_mul = partial(be.tmul, lr)

            cop1.parameter_update(gu.grad_apply(lr_mul, grad))
            cop1_state, cop1_GFE = cop1.compute_StateTAP(use_GD=False, tol=1e-7, max_iters=50)

            cop2 = deepcopy(rbm)
            cop2.parameter_update(gu.grad_apply(lr_mul, random_grad))
            cop2_state, cop2_GFE = cop2.compute_StateTAP(use_GD=False, tol=1e-7, max_iters=50)

            regress = cop2_GFE - cop1_GFE < 0

            if regress:
                if lr < 1e-6:
                    assert False, \
                    "TAP FE gradient is not working properly for Gaussian models"
                    break
                else:
                    lr *= 0.5
            else:
                break

def test_gaussian_GFE_derivatives_self_consistent():
    num_units = 5

    layer_1 = layers.GaussianLayer(num_units)
    layer_2 = layers.BernoulliLayer(num_units)

    rbm = BoltzmannMachine([layer_1, layer_2])
    for i in range(len(rbm.connections)):
        rbm.connections[i].weights.params.matrix[:] = \
        0.01 * be.randn(rbm.connections[i].shape)

    for lay in rbm.layers:
        lay.params.loc[:] = be.rand_like(lay.params.loc)

    state, cop1_GFE = rbm._compute_StateTAP_self_consistent()
    grad = rbm._grad_gibbs_free_energy(state)
    gu.grad_normalize_(grad)

    for i in range(100):
        lr = 0.1
        gogogo = True
        random_grad = gu.random_grad(rbm)
        gu.grad_normalize_(random_grad)
        while gogogo:
            cop1 = deepcopy(rbm)
            lr_mul = partial(be.tmul, lr)

            cop1.parameter_update(gu.grad_apply(lr_mul, grad))
            cop1_state, cop1_GFE = cop1._compute_StateTAP_self_consistent()

            cop2 = deepcopy(rbm)
            cop2.parameter_update(gu.grad_apply(lr_mul, random_grad))
            cop2_state, cop2_GFE = cop2._compute_StateTAP_self_consistent()

            regress = cop2_GFE - cop1_GFE < 0.0

            if regress:
                if lr < 1e-6:
                    assert False, \
                    "TAP FE gradient is not working properly for Gaussian models"
                    break
                else:
                    lr *= 0.5
            else:
                break

def test_gaussian_Compute_StateTAP_self_consistent():
    num_units = 10

    layer_1 = layers.GaussianLayer(num_units)
    layer_2 = layers.BernoulliLayer(num_units)

    rbm = BoltzmannMachine([layer_1, layer_2])
    for i in range(len(rbm.connections)):
        rbm.connections[i].weights.params.matrix[:] = \
        0.01 * be.randn(rbm.connections[i].shape)

    for lay in rbm.layers:
        lay.params.loc[:] = be.rand_like(lay.params.loc)

    for i in range(100):
        random_state = StateTAP.from_model_rand(rbm)
        GFE = rbm.gibbs_free_energy(random_state.cumulants)
        _,min_GFE = rbm._compute_StateTAP_self_consistent(seed=random_state)

        if GFE - min_GFE < 0.0:
            assert False, \
                "compute_StateTAP_self_consistent is not reducing the GFE"

def test_gaussian_Compute_StateTAP_GD():
    num_units = 10

    layer_1 = layers.GaussianLayer(num_units)
    layer_2 = layers.BernoulliLayer(num_units)

    rbm = BoltzmannMachine([layer_1, layer_2])
    for i in range(len(rbm.connections)):
        rbm.connections[i].weights.params.matrix[:] = \
        0.01 * be.randn(rbm.connections[i].shape)

    for lay in rbm.layers:
        lay.params.loc[:] = be.rand_like(lay.params.loc)

    for i in range(100):
        random_state = StateTAP.from_model_rand(rbm)
        GFE = rbm.gibbs_free_energy(random_state.cumulants)
        _,min_GFE = rbm._compute_StateTAP_GD(seed=random_state)

        if GFE - min_GFE < 0.0:
            assert False, \
                "compute_StateTAP_self_consistent is not reducing the GFE"


def test_gaussian_conditional_params():
    num_visible_units = 100
    num_hidden_units = 50
    batch_size = 25

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.GaussianLayer(num_visible_units)
    hid_layer = layers.GaussianLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # randomly set the intrinsic model parameters
    a = be.randn((num_visible_units,))
    b = be.randn((num_hidden_units,))
    log_var_a = 0.1 * be.randn((num_visible_units,))
    log_var_b = 0.1 * be.randn((num_hidden_units,))
    W = be.randn((num_visible_units, num_hidden_units))

    rbm.layers[0].params.loc[:] = a
    rbm.layers[1].params.loc[:] = b
    rbm.layers[0].params.log_var[:] = log_var_a
    rbm.layers[1].params.log_var[:] = log_var_b
    rbm.connections[0].weights.params.matrix[:] = W

    # generate a random batch of data
    vdata = rbm.layers[0].random((batch_size, num_visible_units))
    hdata = rbm.layers[1].random((batch_size, num_hidden_units))

    # compute the variance
    visible_var = be.exp(log_var_a)
    hidden_var = be.exp(log_var_b)

    # rescale the data
    vdata_scaled = vdata / visible_var
    hdata_scaled = hdata / hidden_var

    # test rescale
    assert be.allclose(vdata_scaled, rbm.layers[0].rescale(vdata)),\
    "visible rescale wrong in gaussian-gaussian rbm"

    assert be.allclose(hdata_scaled, rbm.layers[1].rescale(hdata)),\
    "hidden rescale wrong in gaussian-gaussian rbm"

    # compute the mean
    hidden_mean = be.dot(vdata_scaled, W) # (batch_size, num_hidden_units)
    hidden_mean += b

    visible_mean = be.dot(hdata_scaled, be.transpose(W)) # (batch_size, num_hidden_units)
    visible_mean += a

    # update the conditional parameters using the layer functions
    vis_mean_func, vis_var_func = rbm.layers[0].conditional_params(
        [hdata_scaled], [rbm.connections[0].W(trans=True)])
    hid_mean_func, hid_var_func = rbm.layers[1].conditional_params(
        [vdata_scaled], [rbm.connections[0].W()])

    assert be.allclose(visible_var, vis_var_func),\
    "visible variance wrong in gaussian-gaussian rbm"

    assert be.allclose(hidden_var, hid_var_func),\
    "hidden variance wrong in gaussian-gaussian rbm"

    assert be.allclose(visible_mean, vis_mean_func),\
    "visible mean wrong in gaussian-gaussian rbm"

    assert be.allclose(hidden_mean, hid_mean_func),\
    "hidden mean wrong in gaussian-gaussian rbm"


def test_gaussian_derivatives():
    num_visible_units = 100
    num_hidden_units = 50
    batch_size = 25

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.GaussianLayer(num_visible_units)
    hid_layer = layers.GaussianLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # randomly set the intrinsic model parameters
    a = be.randn((num_visible_units,))
    b = be.randn((num_hidden_units,))
    log_var_a = 0.1 * be.randn((num_visible_units,))
    log_var_b = 0.1 * be.randn((num_hidden_units,))
    W = be.randn((num_visible_units, num_hidden_units))

    rbm.layers[0].params.loc[:] = a
    rbm.layers[1].params.loc[:] = b
    rbm.layers[0].params.log_var[:] = log_var_a
    rbm.layers[1].params.log_var[:] = log_var_b
    rbm.connections[0].weights.params.matrix[:] = W

    # generate a random batch of data
    vdata = rbm.layers[0].random((batch_size, num_visible_units))
    visible_var = be.exp(log_var_a)
    vdata_scaled = vdata / visible_var

    # compute the mean of the hidden layer
    hid_mean = rbm.layers[1].conditional_mean(
        [vdata_scaled], [rbm.connections[0].W()])
    hidden_var = be.exp(log_var_b)
    hid_mean_scaled = rbm.layers[1].rescale(hid_mean)

    # compute the derivatives
    d_vis_loc = be.mean((a-vdata)/visible_var, axis=0)
    d_vis_logvar = -0.5 * be.mean(be.square(be.subtract(a, vdata)), axis=0)
    d_vis_logvar += be.batch_quadratic(hid_mean_scaled, be.transpose(W), vdata,
                                 axis=0) / len(vdata)
    d_vis_logvar /= visible_var

    d_hid_loc = be.mean((b-hid_mean)/hidden_var, axis=0)

    d_hid_logvar = -0.5 * be.mean(be.square(hid_mean - b), axis=0)
    d_hid_logvar += be.batch_quadratic(vdata_scaled, W, hid_mean,
                                 axis=0) / len(hid_mean)
    d_hid_logvar /= hidden_var

    d_W = -be.batch_outer(vdata_scaled, hid_mean_scaled) / len(vdata_scaled)

    # compute the derivatives using the layer functions
    vis_derivs = rbm.layers[0].derivatives(vdata, [hid_mean_scaled],
                                            [rbm.connections[0].W(trans=True)])

    hid_derivs = rbm.layers[1].derivatives(hid_mean, [vdata_scaled],
                                           [rbm.connections[0].W()])

    weight_derivs = rbm.connections[0].weights.derivatives(vdata_scaled, hid_mean_scaled)

    # compute simple weighted derivatives using the layer functions
    scale = 2
    scale_func = partial(be.multiply, be.float_scalar(scale))
    vis_derivs_scaled = rbm.layers[0].derivatives(vdata, [hid_mean_scaled],
                        [rbm.connections[0].W(trans=True)], weighting_function=scale_func)

    hid_derivs_scaled = rbm.layers[1].derivatives(hid_mean, [vdata_scaled],
                        [rbm.connections[0].W()], weighting_function=scale_func)

    weight_derivs_scaled = rbm.connections[0].weights.derivatives(vdata_scaled, hid_mean_scaled,
                                                weighting_function=scale_func)

    assert be.allclose(d_vis_loc, vis_derivs[0].loc), \
    "derivative of visible loc wrong in gaussian-gaussian rbm"

    assert be.allclose(d_hid_loc, hid_derivs[0].loc), \
    "derivative of hidden loc wrong in gaussian-gaussian rbm"

    assert be.allclose(d_vis_logvar, vis_derivs[0].log_var, rtol=1e-05, atol=1e-01), \
    "derivative of visible log_var wrong in gaussian-gaussian rbm"

    assert be.allclose(d_hid_logvar, hid_derivs[0].log_var, rtol=1e-05, atol=1e-01), \
    "derivative of hidden log_var wrong in gaussian-gaussian rbm"

    assert be.allclose(d_W, weight_derivs[0].matrix), \
    "derivative of weights wrong in gaussian-gaussian rbm"

    assert be.allclose(scale * d_vis_loc, vis_derivs_scaled[0].loc), \
    "weighted derivative of visible loc wrong in gaussian-gaussian rbm"

    assert be.allclose(scale * d_hid_loc, hid_derivs_scaled[0].loc), \
    "weighted derivative of hidden loc wrong in gaussian-gaussian rbm"

    assert be.allclose(scale * d_vis_logvar, vis_derivs_scaled[0].log_var, rtol=1e-05, atol=1e-01), \
    "weighted derivative of visible log_var wrong in gaussian-gaussian rbm"

    assert be.allclose(scale * d_hid_logvar, hid_derivs_scaled[0].log_var, rtol=1e-05, atol=1e-01), \
    "weighted derivative of hidden log_var wrong in gaussian-gaussian rbm"

    assert be.allclose(scale * d_W, weight_derivs_scaled[0].matrix), \
    "weighted derivative of weights wrong in gaussian-gaussian rbm"


def test_onehot_conditional_params():
    num_visible_units = 100
    num_hidden_units = 50
    batch_size = 25

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.OneHotLayer(num_visible_units)
    hid_layer = layers.OneHotLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # randomly set the intrinsic model parameters
    a = be.randn((num_visible_units,))
    b = be.randn((num_hidden_units,))
    W = be.randn((num_visible_units, num_hidden_units))

    rbm.layers[0].params.loc[:] = a
    rbm.layers[1].params.loc[:] = b
    rbm.connections[0].weights.params.matrix[:] = W

    # generate a random batch of data
    vdata = rbm.layers[0].random((batch_size, num_visible_units))
    hdata = rbm.layers[1].random((batch_size, num_hidden_units))

    # compute conditional parameters
    hidden_field = be.dot(vdata, W) # (batch_size, num_hidden_units)
    hidden_field += b

    visible_field = be.dot(hdata, be.transpose(W)) # (batch_size, num_visible_units)
    visible_field += a

    # compute conditional parameters with layer funcitons
    hidden_field_layer = rbm.layers[1].conditional_params(
        [vdata], [rbm.connections[0].W()])
    visible_field_layer = rbm.layers[0].conditional_params(
        [hdata], [rbm.connections[0].W(trans=True)])

    assert be.allclose(hidden_field, hidden_field_layer), \
    "hidden field wrong in onehot-onehot rbm"

    assert be.allclose(visible_field, visible_field_layer), \
    "visible field wrong in onehot-onehot rbm"


def test_onehot_derivatives():
    num_visible_units = 100
    num_hidden_units = 50
    batch_size = 25

    # set a seed for the random number generator
    be.set_seed()

    # set up some layer and model objects
    vis_layer = layers.OneHotLayer(num_visible_units)
    hid_layer = layers.OneHotLayer(num_hidden_units)
    rbm = BoltzmannMachine([vis_layer, hid_layer])

    # randomly set the intrinsic model parameters
    a = be.randn((num_visible_units,))
    b = be.randn((num_hidden_units,))
    W = be.randn((num_visible_units, num_hidden_units))

    rbm.layers[0].params.loc[:] = a
    rbm.layers[1].params.loc[:] = b
    rbm.connections[0].weights.params.matrix[:] = W

    # generate a random batch of data
    vdata = rbm.layers[0].random((batch_size, num_visible_units))
    vdata_scaled = rbm.layers[0].rescale(vdata)

    # compute the conditional mean of the hidden layer
    hid_mean = rbm.layers[1].conditional_mean([vdata], [rbm.connections[0].W()])
    hid_mean_scaled = rbm.layers[1].rescale(hid_mean)

    # compute the derivatives
    d_visible_loc = -be.mean(vdata, axis=0)
    d_hidden_loc = -be.mean(hid_mean_scaled, axis=0)
    d_W = -be.batch_outer(vdata, hid_mean_scaled) / len(vdata)

    # compute the derivatives using the layer functions
    vis_derivs = rbm.layers[0].derivatives(vdata, [hid_mean_scaled],
                                            [rbm.connections[0].W(trans=True)])

    hid_derivs = rbm.layers[1].derivatives(hid_mean, [vdata_scaled],
                                          [rbm.connections[0].W()])

    weight_derivs = rbm.connections[0].weights.derivatives(vdata, hid_mean_scaled)

    # compute simple weighted derivatives using the layer functions
    scale = 2
    scale_func = partial(be.multiply, be.float_scalar(scale))
    vis_derivs_scaled = rbm.layers[0].derivatives(vdata, [hid_mean_scaled],
                        [rbm.connections[0].W(trans=True)], weighting_function=scale_func)

    hid_derivs_scaled = rbm.layers[1].derivatives(hid_mean, [vdata_scaled],
                          [rbm.connections[0].W()], weighting_function=scale_func)

    weight_derivs_scaled = rbm.connections[0].weights.derivatives(vdata, hid_mean_scaled,
                                                weighting_function=scale_func)

    assert be.allclose(d_visible_loc, vis_derivs[0].loc), \
    "derivative of visible loc wrong in onehot-onehot rbm"

    assert be.allclose(d_hidden_loc, hid_derivs[0].loc), \
    "derivative of hidden loc wrong in onehot-onehot rbm"

    assert be.allclose(d_W, weight_derivs[0].matrix), \
    "derivative of weights wrong in onehot-onehot rbm"

    assert be.allclose(scale * d_visible_loc, vis_derivs_scaled[0].loc), \
    "weighted derivative of visible loc wrong in onehot-onehot rbm"

    assert be.allclose(scale * d_hidden_loc, hid_derivs_scaled[0].loc), \
    "weighted derivative of hidden loc wrong in onehot-onehot rbm"

    assert be.allclose(scale * d_W, weight_derivs_scaled[0].matrix), \
    "weighted derivative of weights wrong in onehot-onehot rbm"

if __name__ == "__main__":
    pytest.main([__file__])



import sys

from . import backends as be


class Transformation(object):

    def __init__(self, function=be.do_nothing, args=None, kwargs=None):
        """
        Create a transformation that operates on a list of tensors.

        Args:
            function (optional; callable)
            args (optional; List)
            kwargs (optional; Dict)

        Returns:
            Transformation

        """
        self.name = function.__name__
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}

    def _closure(self):
        """
        Create a callable function with the arguments and keyword arguments
        already in place.

        Args:
            None

        Returns:
            callable

        """
        def partial(tensor):
            return self.function(tensor, *self.args, **self.kwargs)
        return partial

    def compute(self, tensor):
        """
        Apply the transformation to a single tensor.

        Args:
            tensor

        Returns:
            tensor

        """
        return self._closure()(tensor)

    def get_config(self):
        """
        Get the configuration of a transformation.

        Args:
            None

        Returns:
            Dict

        """
        return {'name': self.name,
                'args': self.args if len(self.args) > 0 else None,
                'kwargs': self.kwargs if len(self.kwargs) > 0 else None}

    @classmethod
    def from_config(cls, config):
        """
        Create a transformation from a configuration dictionary.

        Args:
            config (Dict)

        Returns:
            Transformation

        """
        function = getattr(sys.modules[__name__], config["name"])
        return cls(function, config['args'], config['kwargs'])


def scale(tensor, denominator=1):
    """
    Rescale the values in a tensor by the denominator.

    Args:
        tensor (tensor (num_samples, num_units))
        denominator (optional; float)

    Returns:
        tensor (tensor (num_samples, num_units))

    """
    return tensor/denominator


def l2_normalize(tensor):
    """
    Divide the rows of a tensor by their L2 norms.

    Args:
        tensor (tensor (num_samples, num_units))

    Returns:
        tensor (tensor (num_samples, num_units))

    """
    return be.divide(be.norm(tensor, axis=1, keepdims=True), tensor)


def l1_normalize(tensor):
    """
    Divide the rows of a tensor by their L1 norms.

    Args:
        tensor (tensor (num_samples, num_units))

    Returns:
        tensor (tensor (num_samples, num_units))

    """
    return be.divide(be.tsum(tensor, axis=1, keepdims=True), tensor)


def binarize_color(tensor):
    """
    Scales an int8 "color" value to [0, 1].

    Args:
        tensor (tensor (num_samples, num_units))

    Returns:
        tensor (tensor (num_samples, num_units))

    """
    return be.float_tensor(be.tround(tensor/255))


def one_hot(data, category_list):
    """
    Convert a categorical variable into a one-hot code.

    Args:
        data (tensor (num_samples, 1)): a column of the data matrix that is categorical
        category_list: the list of categories

    Returns:
        one-hot encoded data (tensor (num_samples, num_categories))

    """
    units = be.zeros((len(data), len(category_list)))
    on_units = be.long_tensor(list(map(category_list.index, be.flatten(data))))
    be.scatter_(units, on_units, 1.)
    return units



import numpy as np

from paysage import backends as be
from paysage import preprocess as pre

import pytest

tensors = [be.rand((100, 10)) for _ in range(8)]

def compare_lists(a, b):
    return all([be.allclose(ai, bi) for ai, bi in zip(a, b)])

def test_scale():
    # test function
    result_pre = [pre.scale(tensor, 2) for tensor in tensors]
    result_ref = [0.5 * tensor for tensor in tensors]
    assert compare_lists(result_pre, result_ref)
    # test transform
    transformer = pre.Transformation(pre.scale, kwargs={'denominator': 2})
    result_pre_2 = [transformer.compute(tensor) for tensor in tensors]
    assert compare_lists(result_pre, result_pre_2)

def test_l2_normalize():
    result_pre = [be.norm(pre.l2_normalize(tensor), axis=1) for tensor in tensors]
    result_ref = [be.ones((len(tensor),)) for tensor in tensors]
    assert compare_lists(result_pre, result_ref)

def test_l1_normalize():
    result_pre = [be.tsum(pre.l1_normalize(tensor), axis=1) for tensor in tensors]
    result_ref = [be.ones((len(tensor),)) for tensor in tensors]
    assert compare_lists(result_pre, result_ref)

def test_binarize_color():
    result_pre = [pre.binarize_color(pre.scale(tensor, 1/255)) for tensor in tensors]
    result_ref = [be.float_tensor(be.tround(tensor)) for tensor in tensors]
    assert compare_lists(result_pre, result_ref)

def test_one_hot():
    categories = range(10)
    labels = be.unsqueeze(be.long_tensor(np.arange(100) // 10), 1)
    hots = pre.one_hot(labels, categories)
    hots_ref = be.zeros((len(labels), len(categories)))
    be.scatter_(hots_ref, be.long_tensor(np.arange(100) // 10), 1)
    assert be.allclose(hots, hots_ref)

def test_transformation_config():
    transformer = pre.Transformation(pre.scale, kwargs={'denominator': 2})
    transformer_result = [transformer.compute(tensor) for tensor in tensors]
    config = transformer.get_config()
    transformer_from_config = pre.Transformation.from_config(config)
    config_result = [transformer_from_config.compute(tensor) for tensor in tensors]
    assert compare_lists(transformer_result, config_result)

if __name__ == "__main__":
    pytest.main([__file__])



from copy import deepcopy

from cytoolz import identity, partial

from . import backends as be
from .models import gradient_util as gu
from . import schedules

# ----- CLASSES ----- #

class GradientMemory(object):
    """
    Many optimizers like RMSProp or ADAM keep track of moving averages
    of the gradients. This class computes the first two moments of the
    gradients as running averages.

    """

    def __init__(self, mean_weight=0.9, mean_square_weight=0.0):
        """
        Create a gradient memory object to keep track of the first two
        moments of the gradient.

        Args:
            mean_weight (float \in (0,1); optional):
                how strongly to weight the previous gradient
            mean_square_weight (float \in (0,1); optional)
                how strongly to weight the square of the previous gradient

        Returns:
            GradientMemory

        """
        self.mean_weight = be.float_scalar(mean_weight)
        self.mean_square_weight = be.float_scalar(mean_square_weight)

        self.mean_gradient = None
        self.mean_square_gradient = None

        self.mixer_ = partial(be.mix_, self.mean_weight)
        self.square_mixer_ = partial(be.square_mix_, self.mean_square_weight)


    def reset(self):
        """
        Reset the accululated mean and mean square gradients.

        Notes:
            Modifies mean_gradient and mean_square_gradient in place.

        Args:
            None

        Returns:
            None

        """
        self.mean_gradient = None
        self.mean_square_gradient = None

    def update_mean(self, grad):
        """
        Update the running average of the model gradients.

        Args:
            grad (a Gradient object)

        Returns:
            None

        """
        if self.mean_gradient is None:
            self.mean_gradient = gu.grad_apply(identity, grad)
        else:
            gu.grad_mapzip_(self.mixer_, self.mean_gradient, grad)

    def update_mean_square(self, grad):
        """
        Update the running average of the squared model gradients.

        Args:
            grad (a Gradient object)

        Returns:
            None

        """
        if self.mean_square_gradient is None:
            self.mean_square_gradient = gu.grad_apply(be.square, grad)
        else:
            gu.grad_mapzip_(self.square_mixer_, self.mean_square_gradient, grad)

    def update(self, grad):
        """
        Update the running average of the model gradients and the running
        average of the squared model gradients.

        Notes:
            Modifies mean_weight and mean_square_weight attributes in place.

        Args:
            grad (a Gradient object)

        Returns:
            None

        """
        if self.mean_weight:
            self.update_mean(grad)
        if self.mean_square_weight:
            self.update_mean_square(grad)

    def normalize(self, grad, unbiased=False):
        """
        Divide grad by the square root of the mean square gradient.

        Notes:
            A running average is biased due to autoregressive correlations
            between adjacent timepoints. The bias can be corrected by
            dividing the results by appropriate weights that reflect
            the degree of autocorrelation.

            Acts like the identity function if mean_square_weight = 0.

        Args:
            grad (a Gradient object)
            unbiased (bool): whether to unbias the estimates

        Returns:
            normalized Gradient object

        """
        if not self.mean_square_gradient:
            return grad

        if unbiased:
            mean_norm = be.float_scalar(1 - self.mean_weight)
            mean_square_norm = be.float_scalar(1 - self.mean_square_weight)
            def normalizer(mean, mean_square):
                return be.sqrt_div(mean / mean_norm,
                                   mean_square / mean_square_norm)
        else:
            def normalizer(mean, mean_square):
                return be.sqrt_div(mean, mean_square)

        return gu.grad_mapzip(normalizer, grad, self.mean_square_gradient)


class Optimizer(object):
    """Base class for the optimizer methods."""
    def __init__(self,
                 stepsize=schedules.Constant(initial=0.001),
                 tolerance=1e-7):
        """
        Create an optimizer object:

        Args:
            model: a BoltzmannMachine object to optimize
            stepsize (generator; optional): the stepsize schedule
            tolerance (float; optional):
                the gradient magnitude to declar convergence

        Returns:
            Optimizer

        """
        self.stepsize = stepsize
        self.tolerance = tolerance
        self.delta = {}
        self.lr_ = partial(be.tmul_, stepsize)

    def check_convergence(self):
        """
        Check the convergence criterion.

        Args:
            None

        Returns:
            bool: True if converged, False if not
        """
        mag = gu.grad_rms(self.delta)
        return mag <= self.tolerance

    def update_lr(self):
        """
        Update the current value of the stepsize:

        Notes:
            Modifies stepsize attribute in place.

        Args:
            None

        Returns:
            None

        """
        lr = be.float_scalar(next(self.stepsize))
        self.lr_ = partial(be.tmul_, lr)


class Gradient(Optimizer):
    """Vanilla gradient optimizer"""
    def __init__(self,
                 stepsize=schedules.Constant(initial=0.001),
                 tolerance=1e-7):
        """
        Create a gradient descent optimizer.

        Aliases:
            gradient

        Args:
            model: a BoltzmannMachine object to optimize
            stepsize (generator; optional): the stepsize schedule
            tolerance (float; optional):
                the gradient magnitude to declar convergence

        Returns:
            StochasticGradientDescent

        """
        super().__init__(stepsize, tolerance)

    def reset(self):
        """
        Reset the gradient memory (does nothing for vanilla gradient).

        Notes:
            Modifies gradient memory in place.

        Args:
            None

        Returns:
            None

        """
        pass

    def update(self, model, grad):
        """
        Update the model parameters with a gradient step.

        Notes:
            Changes parameters of model in place.

        Args:
            model: a BoltzmannMachine object to optimize
            grad: a Gradient object
            epoch (int): the current epoch

        Returns:
            None

        """
        self.delta = deepcopy(grad)
        gu.grad_apply_(self.lr_, self.delta)
        model.parameter_update(self.delta)


class Momentum(Optimizer):
    """
    Stochastic gradient descent with momentum.
    Qian, N. (1999).
    On the momentum term in gradient descent learning algorithms.
    Neural Networks, 12(1), 145–151

    """
    def __init__(self,
                 stepsize=schedules.Constant(initial=0.001),
                 momentum=0.9,
                 tolerance=1e-7):
        """
        Create a stochastic gradient descent with momentum optimizer.

        Aliases:
            momentum

        Args:
            model: a BoltzmannMachine object to optimize
            stepsize (generator; optional): the stepsize schedule
            momentum (float; optional): the amount of momentum
            tolerance (float; optional):
                the gradient magnitude to declar convergence

        Returns:
            Momentum

        """
        super().__init__(stepsize, tolerance)
        self.memory = GradientMemory(mean_weight=momentum,
                                     mean_square_weight=0)

    def reset(self):
        """
        Reset the gradient memory.

        Notes:
            Modifies gradient memory in place.

        Args:
            None

        Returns:
            None

        """
        self.memory.reset()

    def update(self, model, grad):
        """
        Update the model parameters with a gradient step.

        Notes:
            Changes parameters of model in place.

        Args:
            model: a BoltzmannMachine object to optimize
            grad: a Gradient object
            epoch (int): the current epoch

        Returns:
            None

        """
        self.memory.update(grad)
        self.delta = deepcopy(self.memory.mean_gradient)
        gu.grad_apply_(self.lr_, self.delta)
        model.parameter_update(self.delta)


class RMSProp(Optimizer):
    """
    Stochastic gradient descent with RMSProp.
    Geoffrey Hinton's Coursera Course Lecture 6e

    """
    def __init__(self,
                 stepsize=schedules.Constant(initial=0.001),
                 mean_square_weight=0.9,
                 tolerance=1e-7):
        """
        Create a stochastic gradient descent with RMSProp optimizer.

        Aliases:
            rmsprop

        Args:
            model: a BoltzmannMachine object to optimize
            stepsize (generator; optional): the stepsize schedule
            mean_square_weight (float; optional):
                for computing the running average of the mean-square gradient
            tolerance (float; optional):
                the gradient magnitude to declar convergence

        Returns:
            RMSProp

        """
        super().__init__(stepsize, tolerance)
        self.memory = GradientMemory(mean_weight=0,
                                     mean_square_weight=mean_square_weight)

    def reset(self):
        """
        Reset the gradient memory.

        Notes:
            Modifies gradient memory in place.

        Args:
            None

        Returns:
            None

        """
        self.memory.reset()

    def update(self, model, grad):
        """
        Update the model parameters with a gradient step.

        Notes:
            Changes parameters of model in place.

        Args:
            model: a BoltzmannMachine object to optimize
            grad: a Gradient object
            epoch (int): the current epoch

        Returns:
            None

        """
        self.memory.update(grad)
        self.delta = self.memory.normalize(grad, True)
        gu.grad_apply_(self.lr_, self.delta)
        model.parameter_update(self.delta)


class ADAM(Optimizer):
    """
    Stochastic gradient descent with Adaptive Moment Estimation algorithm.

    Kingma, D. P., & Ba, J. L. (2015).
    Adam: a Method for Stochastic Optimization.
    International Conference on Learning Representations, 1–13.

    """
    def __init__(self,
                 stepsize=schedules.Constant(initial=0.001),
                 mean_weight=0.9,
                 mean_square_weight=0.999,
                 tolerance=1e-7):
        """
        Create a stochastic gradient descent with ADAM optimizer.

        Aliases:
            adam

        Args:
            model: a BoltzmannMachine object to optimize
            stepsize (generator; optional): the stepsize schedule
            mean_weight (float; optional):
                for computing the running average of the mean gradient
            mean_square_weight (float; optional):
                for computing the running average of the mean-square gradient
            tolerance (float; optional):
                the gradient magnitude to declar convergence

        Returns:
            ADAM

        """
        super().__init__(stepsize, tolerance)
        self.memory = GradientMemory(mean_weight=mean_weight,
                                     mean_square_weight=mean_square_weight)

    def reset(self):
        """
        Reset the gradient memory.

        Notes:
            Modifies gradient memory in place.

        Args:
            None

        Returns:
            None

        """
        self.memory.reset()

    def update(self, model, grad):
        """
        Update the model parameters with a gradient step.

        Notes:
            Changes parameters of model in place.

        Args:
            model: a BoltzmannMachine object to optimize
            grad: a Gradient object
            epoch (int): the current epoch

        Returns:
            None

        """
        self.memory.update(grad)
        self.delta = self.memory.normalize(self.memory.mean_gradient, True)
        gu.grad_apply_(self.lr_, self.delta)
        model.parameter_update(self.delta)


# ----- ALIASES ----- #

gradient = Gradient
momentum = Momentum
rmsprop = RMSProp
adam = ADAM



from . import backends as be
from . import schedules
from .models import state as model_state

class AutoregressiveGammaSampler(object):
    """Sampler from an autoregressive Gamma process."""
    def __init__(self, beta_momentum=0.9, beta_std=0.6,
                 schedule=schedules.Constant(initial=1.0)):
        """
        Create an autoregressive gamma sampler.
        Can be used to sample inverse temperatures for MC sampling.

        Args:
            beta_momentum (float in [0,1]; optional): autoregressive coefficient
                the inverse temperature, beta.
            beta_std (float >= 0; optional): the standard deviation of the
                inverse temperature, beta.
            schedule (generator; optional)

        Returns:
            An AutoregressiveGammaSampler instance.

        """
        from numpy.random import gamma, poisson
        self.gamma = gamma
        self.poisson = poisson

        self.schedule = schedule
        self.set_std(beta_std, beta_momentum)

    def set_std(self, std, momentum=0.9):
        """
        Set the parameters based off the standard deviation.

        Notes:
            Modifies many layer attributes in place!

        Args:
            std (float)

        Returns:
            None

        """
        self.std = std
        self.var = self.std**2
        self.use_driven = (self.std > 0)

        if self.use_driven:
            self.phi = momentum # autocorrelation
            self.nu = 1 / self.var # location parameter
            self.c = (1-self.phi) * self.var # scale parameter

        self.beta = None
        self.has_beta = False
        self.schedule.reset()

    def set_schedule(self, value):
        """
        Change the value of the learning rate schedule.

        Notes:
            Modifies the schedule.value attribute in place!

        Args:
            value (float)

        Returns:
            None

        """
        self.schedule.set_value(value)

    def _anneal(self):
        """
        Get the next value from the learning rate schedule and update
        the parameters of the Gamma generating process.

        Args:
            None

        Return:
            nu (float), c (float)

        """
        t = be.EPSILON + next(self.schedule)
        return self.nu / t, self.c * t

    def update_beta(self, num_samples):
        """
        Update beta with an autoregressive Gamma process.

        beta_0 ~ Gamma(nu,c/(1-phi)) = Gamma(nu, var)
        h_t ~ Possion( phi/c * h_{t-1})
        beta_t ~ Gamma(nu + z_t, c)

        Achieves a stationary distribution with mean 1 and variance var:
        Gamma(nu, var) = Gamma(1/var, var)

        Notes:
            Modifies the folling attributes in place:
                has_beta, beta_shape, beta

        Args:
            num_samples (int): the number of samples to generate for beta

        Returns:
            None

        """
        if self.use_driven:
            nu, c = self._anneal()
            if not self.has_beta or self.beta_shape[0] != num_samples:
                self.has_beta = True
                self.beta_shape = (num_samples, 1)
                self.beta = self.gamma(nu, c/(1-self.phi), size=self.beta_shape)
            z = self.poisson(lam=self.beta * self.phi/c)
            self.beta = self.gamma(nu + z, c)

    def get_beta(self):
        """Return beta in the appropriate tensor format."""
        if self.use_driven:
            return be.float_tensor(self.beta)
        return None


class SequentialMC(object):
    """An accelerated sequential Monte Carlo sampler"""
    def __init__(self, model, mcsteps=1, clamped=None, updater='markov_chain',
                 beta_momentum=0.9, beta_std=0.6,
                 schedule=schedules.Constant(initial=1.0)):
        """
        Create a sequential Monte Carlo sampler.

        Args:
            model (BoltzmannMachine)
            mcsteps (int; optional): the number of Monte Carlo steps
            clamped (List[int]; optional): list of layers to clamp
            updater (str; optional): method for updating the state
            beta_momentum (float in [0,1]; optional): autoregressive coefficient
                the inverse temperature of beta
            beta_std (float >= 0; optional): the standard deviation of the
                inverse temperature beta
            schedule (generator; optional)

        Returns:
            SequentialMC

        """
        self.model = model
        self.state = None
        self.update_method = updater
        self.updater = getattr(model, updater)
        self.mcsteps = mcsteps

        self.clamped = []
        if clamped is not None:
            self.clamped = clamped

        self.beta_sampler = AutoregressiveGammaSampler(beta_momentum,
                                                       beta_std,
                                                       schedule)

    def set_state(self, state):
        """
        Set the state.

        Notes:
            Modifies the state attribute in place.

        Args:
            state (State): The state of the units.

        Returns:
            None

        """
        self.state = state

    def set_state_from_visible(self, vdata):
        """
        Set the state of the sampler using a sample of visible vectors.

        Notes:
            Modifies the sampler.state attribute in place.

        Args:
            vdata (tensor~(num_samples,num_units)): a visible state

        Returns:
            None

        """
        self.set_state(model_state.State.from_visible(vdata, self.model))

    def reset(self):
        """
        Reset the sampler state.

        Notes:
            Modifies sampler.state attribute in place.

        Args:
            None

        Returns:
            None

        """
        self.state = None
        self.beta_sampler.beta = None

    def update_state(self, steps=None):
        """
        Update the state of the particles.

        Notes:
            Modifies the state attribute in place.
            Calls the beta_sampler.update_beta() method.

        Args:
            steps (int): the number of Monte Carlo steps

        Returns:
            None

        """
        if not self.state:
            raise AttributeError(
                'You must call the initialize(self, array_or_shape)'
                +' method to set the initial state of the Markov Chain')
        STEPS = self.mcsteps if steps is None else steps
        for _ in range(STEPS):
            self.beta_sampler.update_beta(be.shape(self.state[0])[0])
            clamping = self.model.clamped_sampling
            self.model.set_clamped_sampling(self.clamped)
            self.state = self.updater(1, self.state, beta=self.beta_sampler.get_beta())
            self.model.set_clamped_sampling(clamping)

    def state_for_grad(self, target_layer):
        """
        Peform a mean field update of the target layer.

        Args:
            target_layer (int): the layer to update

        Returns:
            state

        """
        layer_list = range(self.model.num_layers)
        clamping = self.model.clamped_sampling
        self.model.set_clamped_sampling([i for i in layer_list if i != target_layer])
        grad_state = self.model.mean_field_iteration(1, self.state)
        self.model.set_clamped_sampling(clamping)
        return grad_state

    @classmethod
    def from_batch(cls, model, batch, **kwargs):
        """
        Create a sampler from a batch object.

        Args:
            model: a BoltzmannMachine object
            batch: a Batch object
            kwargs (optional)

        Returns:
            sampler

        """
        tmp = cls(model, **kwargs)
        tmp.set_state_from_visible(batch.get('train'))
        batch.reset_generator('all')
        return tmp

    @classmethod
    def from_visible(cls, model, vdata, **kwargs):
        """
        Create a sampler initialized from visible data.

        Args:
            model: a BoltzmannMachine object
            vdata: visible data
            kwargs (optional)

        Returns:
            sampler

        """
        tmp = cls(model, **kwargs)
        tmp.set_state_from_visible(vdata)
        return tmp

    @classmethod
    def from_model(cls, model, batch_size, **kwargs):
        """
        Create a sampler from a model object.

        Args:
            model: a BoltzmannMachine object
            batch_size: the batch size
            kwargs (optional)

        Returns:
            sampler

        """
        tmp = cls(model, **kwargs)
        tmp.set_state(model_state.State.from_model_envelope(
                batch_size, model))
        return tmp

    @classmethod
    def generate_fantasy_state(cls, model, batch_size, update_steps,
                               **kwargs):
        """
        Generate fantasy particles from a model.

        Args:
            model: A BoltzmannMachine object.
            batch_size (int): the number of fantasy particles to generate
            update_steps (int): how many monte carlo steps to run
            kwargs: other keyword arguments for SequentialMC

        Returns:
            State

        """
        sampler = cls.from_model(model, batch_size, **kwargs)
        sampler.update_state(update_steps)
        return sampler.state
